from typing import Final

CONNCOMP_SUFFIX = ".unw.conncomp.tif"
CONNCOMP_SUFFIX_ZEROED = ".unw.conncomp.zeroed.tif"
UNW_SUFFIX = ".unw.tif"
UNW_SUFFIX_ZEROED = ".unw.zeroed.tif"

UINT16_MAX: Final = 65535

DEFAULT_CCL_NODATA: Final = UINT16_MAX
DEFAULT_UNW_NODATA: Final = 0
